//-----------------------------------------------------------------------------
// vflt_mmx.h : LPF(MMX)
// Copyright (c) 2005-2008  N099
//-----------------------------------------------------------------------------
#include "vflt.h"

// 3tap LPF
NOA void lpf_v_3_mmx(pix * restrict dst, const pix * restrict src,
                     int h, int w, int s)
{
    int i,j;
    p_m64c src0, src1, src2;
    p_m64r dst0;
    __m64 ms0,ms1;
    static const u64_i16 mc0 = { 2, 2, 2, 2 };

    w = COUNT8(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m64r)dst;
        src1 = (p_m64c)src;
        src0 = (p_m64c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src2 = (p_m64c)(src + s * ((i+1) >= h ? -1 : 1));
        
        for(j=0; j<w; ++j){
            ms0 = _mm_adds_pi16(*(src0+j), *(src2+j));
            ms1 = _mm_slli_pi16(*(src1+j), 1);

            ms0 = _mm_adds_pi16(ms0, mc0.m);
            ms1 = _mm_adds_pi16(ms1, ms0);
            ms1 = _mm_srai_pi16(ms1, 2);
            
            *(dst0+j) = ms1;
        }
        
        dst += s;
        src += s;
    }

    _mm_empty();
}


// 5tap LPF
NOA static void lpf_v_5_mmx(pix * restrict dst, const pix * restrict src,
                            int h, int w, int s)
{
    int i,j;
    p_m64c src0, src1, src2, src3, src4;
    p_m64r dst0;
    __m64 ms0,ms1,ms2,ms3;
    static const u64_i16 mc0 = {  8192, -2606,  8192, -2606 };
    static const u64_i16 mc1 = { 21596,     0, 21596,     0 };
    static const u64_i32 mc2 = {  (1<<14) - 1,  (1<<14) - 1 };
    
    w = COUNT8(w * sizeof(pix));
  
    for(i=0; i<h; i++){
        dst0 = (p_m64r)dst;
        src2 = (p_m64c)src;
        src1 = (p_m64c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src3 = (p_m64c)(src + s * ((i+1) >= h ? -1 : 1));
        src0 = (p_m64c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src4 = (p_m64c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        
        for(j=0; j<w; ++j){
            ms2 = _mm_adds_pi16(*(src0+j), *(src4+j));
            ms1 = _mm_adds_pi16(*(src1+j), *(src3+j));
            ms0 = ms1;
            
            ms1 = _mm_unpackhi_pi16(ms1, ms2);
            ms0 = _mm_unpacklo_pi16(ms0, ms2);
            
            ms1 = _mm_madd_pi16(ms1, mc0.m);
            ms0 = _mm_madd_pi16(ms0, mc0.m);
            
            ms0 = _mm_add_pi32(ms0, mc2.m);
            ms1 = _mm_add_pi32(ms1, mc2.m);

            ms2 = *(src2+j);
            ms3 = ms2;
            ms3 = _mm_unpackhi_pi16(ms3, mc0.m);
            ms2 = _mm_unpacklo_pi16(ms2, mc0.m);
            
            ms2 = _mm_madd_pi16(ms2, mc1.m);
            ms3 = _mm_madd_pi16(ms3, mc1.m);
            
            ms0 = _mm_add_pi32(ms0, ms2);
            ms1 = _mm_add_pi32(ms1, ms3);
            
            ms0 = _mm_srai_pi32(ms0, 15);
            ms1 = _mm_srai_pi32(ms1, 15);
            
            *(dst0+j) = _mm_packs_pi32(ms0, ms1);
        }
        
        dst += s;
        src += s;
    }
}


// 7tap LPF
NOA void lpf_v_7_mmx(pix * restrict dst, const pix * restrict src,
                     int h, int w, int s)
{
    int i,j;
    p_m64c src0, src1, src2, src3, src4, src5, src6;
    p_m64r dst0;
    __m64 ms0,ms1,ms2,ms3,ms4;
    static const u64_i16 mc0 = { -4000,  1315, -4000,  1315 };
    static const u64_i16 mc1 = {  6876, 24386,  6876, 24386 };
    static const u64_i32 mc2 = {  (1<<14) - 1,  (1<<14) - 1 };

    w = COUNT8(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m64r)dst;
        src3 = (p_m64c)src;
        src2 = (p_m64c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src4 = (p_m64c)(src + s * ((i+1) >= h ? -1 : 1));
        src1 = (p_m64c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src5 = (p_m64c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src0 = (p_m64c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src6 = (p_m64c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        
        for(j=0; j<w; ++j){
            ms2 = _mm_adds_pi16(*(src0+j), *(src6+j)); 
            ms1 = _mm_adds_pi16(*(src1+j), *(src5+j));
            ms0 = ms1;
            
            ms1 = _mm_unpackhi_pi16(ms1, ms2);
            ms0 = _mm_unpacklo_pi16(ms0, ms2);
            
            ms0 = _mm_madd_pi16(ms0, mc0.m);
            ms1 = _mm_madd_pi16(ms1, mc0.m);
            
            ms0 = _mm_add_pi32(ms0, mc2.m);
            ms1 = _mm_add_pi32(ms1, mc2.m);
            
            ms4 = _mm_adds_pi16(*(src2+j), *(src4+j));
            ms3 = *(src3+j);
            ms2 = ms4;
            ms4 = _mm_unpackhi_pi16(ms4, ms3);
            ms2 = _mm_unpacklo_pi16(ms2, ms3);
            
            ms4 = _mm_madd_pi16(ms4, mc1.m);
            ms2 = _mm_madd_pi16(ms2, mc1.m);
            
            ms1 = _mm_add_pi32(ms1, ms4);
            ms0 = _mm_add_pi32(ms0, ms2);
            
            ms1 = _mm_srai_pi32(ms1, 15);
            ms0 = _mm_srai_pi32(ms0, 15);
            
            *(dst0+j) = _mm_packs_pi32(ms0, ms1);
        }
        
        dst += s;
        src += s;
    }

    _mm_empty();
}


// 9tap LPF
NOA static void lpf_v_9_mmx(pix * restrict dst, const pix * restrict src,
                            int h, int w, int s)
{
    int i,j;
    p_m64c src0, src1, src2, src3, src4, src5, src6, src7, src8;
    p_m64r dst0;
    __m64 ms0,ms1,ms2,ms3,ms4;
    static const u64_i16 mc0 = {      2285,  -766,      2285,  -766 };
    static const u64_i16 mc1 = {      5907, -4237,      5907, -4237 };
    static const u64_i16 mc2 = { (1<<14)-1, 26390, (1<<14)-1, 26390 };
    static const u64_i16 mc3 = {         1,     1,         1,    1 };
    
    w = COUNT8(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m64r)dst;
        src4 = (p_m64c)src;
        src3 = (p_m64c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src5 = (p_m64c)(src + s * ((i+1) >= h ? -1 : 1));
        src2 = (p_m64c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src6 = (p_m64c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src1 = (p_m64c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src7 = (p_m64c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src0 = (p_m64c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        src8 = (p_m64c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
    
        for(j=0; j<w; ++j){
            ms2 = _mm_adds_pi16(*(src0+j), *(src8+j));
            ms1 = _mm_adds_pi16(*(src1+j), *(src7+j));
            ms0 = ms1;
            
            ms0 = _mm_unpackhi_pi16(ms0, ms2);
            ms1 = _mm_unpacklo_pi16(ms1, ms2);
            ms0 = _mm_madd_pi16(ms0, mc0.m);
            ms1 = _mm_madd_pi16(ms1, mc0.m);

            //
            ms4 = _mm_adds_pi16(*(src2+j), *(src6+j));
            ms3 = _mm_adds_pi16(*(src3+j), *(src5+j));
            ms2 = ms3;
            
            ms2 = _mm_unpackhi_pi16(ms2, ms4);
            ms3 = _mm_unpacklo_pi16(ms3, ms4);
            ms2 = _mm_madd_pi16(ms2, mc1.m);
            ms3 = _mm_madd_pi16(ms3, mc1.m);
            
            //
            ms0 = _mm_add_pi32(ms0, ms2);
            ms1 = _mm_add_pi32(ms1, ms3);
            
            //
            ms4 = *(src4+j);
            ms2 = ms3 = mc3.m;
            
            ms2 = _mm_unpackhi_pi16(ms2, ms4);
            ms3 = _mm_unpacklo_pi16(ms3, ms4);
            ms2 = _mm_madd_pi16(ms2, mc2.m);
            ms3 = _mm_madd_pi16(ms3, mc2.m);
            
            ms0 = _mm_add_pi32(ms0, ms2);
            ms1 = _mm_add_pi32(ms1, ms3);
            
            ms0 = _mm_srai_pi32(ms0, 15);
            ms1 = _mm_srai_pi32(ms1, 15);
            
            *(dst0+j) = _mm_packs_pi32(ms1, ms0);
        }
        
        dst += s;
        src += s;
    }
}


// 11tap LPF
NOA static void lpf_v_11_mmx(pix * restrict dst, const pix * restrict src,
                             int h, int w, int s)
{
    int i,j;
    p_m64c src0, src1, src2, src3, src4, src5;
    p_m64c src6, src7, src8, src9, srca;
    p_m64r dst0;
    __m64 ms0,ms1,ms2,ms3,ms4;
    static const u64_i16 mc0 = { -1442,   518, -1442,   518 };
    static const u64_i16 mc1 = { -3993,  2693, -3993,  2693 };
    static const u64_i16 mc2 = {  4980, 27256,  4980, 27256 };
    static const u64_i32 mc3 = {    (1<<14)-1,    (1<<14)-1 };
    
    w = COUNT8(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m64r)dst;
        src5 = (p_m64c)src;
        src4 = (p_m64c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src6 = (p_m64c)(src + s * ((i+1) >= h ? -1 : 1));
        src3 = (p_m64c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src7 = (p_m64c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src2 = (p_m64c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src8 = (p_m64c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src1 = (p_m64c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        src9 = (p_m64c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src0 = (p_m64c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srca = (p_m64c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        
        for(j=0; j<w; ++j){
            //
            ms2 = _mm_adds_pi16(*(src0+j), *(srca+j));
            ms1 = _mm_adds_pi16(*(src1+j), *(src9+j));
            ms0 = ms1;
            
            ms0 = _mm_unpackhi_pi16(ms0, ms2);
            ms1 = _mm_unpacklo_pi16(ms1, ms2);
            ms0 = _mm_madd_pi16(ms0, mc0.m);
            ms1 = _mm_madd_pi16(ms1, mc0.m);
            
            ms0 = _mm_add_pi32(ms0, mc3.m);
            ms1 = _mm_add_pi32(ms1, mc3.m);
            
            //
            ms4 = _mm_adds_pi16(*(src2+j), *(src8+j));
            ms3 = _mm_adds_pi16(*(src3+j), *(src7+j));
            ms2 = ms3;
            
            ms2 = _mm_unpackhi_pi16(ms2, ms4);
            ms3 = _mm_unpacklo_pi16(ms3, ms4);
            ms2 = _mm_madd_pi16(ms2, mc1.m);
            ms3 = _mm_madd_pi16(ms3, mc1.m);
            
            ms0 = _mm_add_pi32(ms0, ms2);
            ms1 = _mm_add_pi32(ms1, ms3);
            
            //
            ms3 = _mm_adds_pi16(*(src4+j), *(src6+j));
            ms4 = *(src5+j);
            ms2 = ms3;
            ms2 = _mm_unpackhi_pi16(ms2, ms4);
            ms3 = _mm_unpacklo_pi16(ms3, ms4);
            
            ms2 = _mm_madd_pi16(ms2, mc2.m);
            ms3 = _mm_madd_pi16(ms3, mc2.m);
            
            ms0 = _mm_add_pi32(ms0, ms2);
            ms1 = _mm_add_pi32(ms1, ms3);
            
            ms0 = _mm_srai_pi32(ms0, 15);
            ms1 = _mm_srai_pi32(ms1, 15);
            
            *(dst0+j) = _mm_packs_pi32(ms1, ms0);
        }
        
        dst += s;
        src += s;
    }
}


// 13tap LPF
NOA static void lpf_v_13_mmx(pix * restrict dst, const pix * restrict src,
                             int h, int w, int s)
{
    int i,j;
    p_m64c src0, src1, src2, src3, src4, src5, src6;
    p_m64c src7, src8, src9, srca, srcb, srcc;
    p_m64r dst0;
    __m64 ms0,ms1,ms2,ms3,ms4;
    static const u64_i16 mc0 = { 981, -363, 981, -363 };
    static const u64_i16 mc1 = { 2822, -1841, 2822, -1841 };
    static const u64_i16 mc2 = { 4389, -3737, 4389, -3737 };
    static const u64_i16 mc3 = { (1<<14)-1, 28266, (1<<14)-1, 28266 };
    static const u64_i16 mc4 = { 1, 1, 1, 1 };
    
    w = COUNT8(w * sizeof(pix));

    for(i=0; i<h; i++){
        dst0 = (p_m64r)dst;
        src6 = (p_m64c)src;
        src5 = (p_m64c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src7 = (p_m64c)(src + s * ((i+1) >= h ? -1 : 1));
        src4 = (p_m64c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src8 = (p_m64c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src3 = (p_m64c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src9 = (p_m64c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src2 = (p_m64c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        srca = (p_m64c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src1 = (p_m64c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srcb = (p_m64c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        src0 = (p_m64c)(src + s * ((i-6) >= 0 ? -6 : 6 - 2*i));
        srcc = (p_m64c)(src + s * ((i+6) >= h ? 2*(h-1-i)-6 : 6));
        
        for(j=0; j<w; ++j){
            ms2 = _mm_adds_pi16(*(src0+j), *(srcc+j));
            ms1 = _mm_adds_pi16(*(src1+j), *(srcb+j));
            ms0 = ms1;
            
            ms0 = _mm_unpackhi_pi16(ms0, ms2);
            ms1 = _mm_unpacklo_pi16(ms1, ms2);
            
            ms0 = _mm_madd_pi16(ms0, mc0.m);
            ms1 = _mm_madd_pi16(ms1, mc0.m);
            
            //
            ms4 = _mm_adds_pi16(*(src2+j), *(srca+j));
            ms3 = _mm_adds_pi16(*(src3+j), *(src9+j));
            ms2 = ms3;
            
            ms2 = _mm_unpackhi_pi16(ms2, ms4);
            ms3 = _mm_unpacklo_pi16(ms3, ms4);
            
            ms2 = _mm_madd_pi16(ms2, mc1.m);
            ms3 = _mm_madd_pi16(ms3, mc1.m);
            
            ms0 = _mm_add_pi32(ms0, ms2);
            ms1 = _mm_add_pi32(ms1, ms3);
            
            //
            ms4 = _mm_adds_pi16(*(src4+j), *(src8+j));
            ms3 = _mm_adds_pi16(*(src5+j), *(src7+j));
            ms2 = ms3;
            
            ms2 = _mm_unpackhi_pi16(ms2, ms4);
            ms3 = _mm_unpacklo_pi16(ms3, ms4);
            
            ms2 = _mm_madd_pi16(ms2, mc2.m);
            ms3 = _mm_madd_pi16(ms3, mc2.m);
            
            ms0 = _mm_add_pi32(ms0, ms2);
            ms1 = _mm_add_pi32(ms1, ms3);
            
            //
            ms4 = *(src6+j);
            ms3 = ms2 = mc4.m;
            
            ms2 = _mm_unpackhi_pi16(ms2, ms4);
            ms3 = _mm_unpacklo_pi16(ms3, ms4);
            
            ms2 = _mm_madd_pi16(ms2, mc3.m);
            ms3 = _mm_madd_pi16(ms3, mc3.m);
            
            ms0 = _mm_add_pi32(ms0, ms2);
            ms1 = _mm_add_pi32(ms1, ms3);
            
            ms0 = _mm_srai_pi32(ms0, 15);
            ms1 = _mm_srai_pi32(ms1, 15);
            
            *(dst0+j) = _mm_packs_pi32(ms1, ms0);
        }
        
        dst += s;
        src += s;
    }
}


// 15tap LPF
NOA void lpf_v_15_mmx(pix * restrict dst, const pix * restrict src,
                      int h, int w, int s)
{
    int i,j;
    p_m64c src0, src1, src2, src3, src4, src5, src6, src7;
    p_m64c src8, src9, srca, srcb, srcc, srcd, srce;
    p_m64r dst0;
    __m64 ms0,ms1,ms2,ms3,ms4;
    static const u64_i16 mc0 = {  -695,  265,  -695,  265 };
    static const u64_i16 mc1 = { -2021, 1299, -2021, 1299 };
    static const u64_i16 mc2 = { -3417, 2766, -3417, 2766 };
    static const u64_i16 mc3 = { 28650, 3862, 28650, 3862 };
    static const u64_i32 mc4 = { (1<<14) - 1, (1<<14) - 1 };

    w = COUNT8(w * sizeof(pix));
    
    for(i=0; i<h; i++){
        dst0 = (p_m64r)dst;
        src7 = (p_m64c)src;
        src6 = (p_m64c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src8 = (p_m64c)(src + s * ((i+1) >= h ? -1 : 1));
        src5 = (p_m64c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src9 = (p_m64c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src4 = (p_m64c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        srca = (p_m64c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src3 = (p_m64c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        srcb = (p_m64c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src2 = (p_m64c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srcc = (p_m64c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        src1 = (p_m64c)(src + s * ((i-6) >= 0 ? -6 : 6 - 2*i));
        srcd = (p_m64c)(src + s * ((i+6) >= h ? 2*(h-1-i)-6 : 6));
        src0 = (p_m64c)(src + s * ((i-7) >= 0 ? -7 : 7 - 2*i));
        srce = (p_m64c)(src + s * ((i+7) >= h ? 2*(h-1-i)-7 : 7));
        
        for(j=0; j<w; ++j){
            ms2 = _mm_adds_pi16(*(src0+j), *(srce+j));
            ms1 = _mm_adds_pi16(*(src1+j), *(srcd+j));
            ms0 = ms1;
            
            ms0 = _mm_unpackhi_pi16(ms0, ms2);
            ms1 = _mm_unpacklo_pi16(ms1, ms2);
            
            ms0 = _mm_madd_pi16(ms0, mc0.m);
            ms1 = _mm_madd_pi16(ms1, mc0.m);
            
            ms0 = _mm_add_pi32(ms0, mc4.m);
            ms1 = _mm_add_pi32(ms1, mc4.m);
            
            //
            ms4 = _mm_adds_pi16(*(src2+j), *(srcc+j));
            ms3 = _mm_adds_pi16(*(src3+j), *(srcb+j));
            ms2 = ms3;
            
            ms2 = _mm_unpackhi_pi16(ms2, ms4);
            ms3 = _mm_unpacklo_pi16(ms3, ms4);

            ms2 = _mm_madd_pi16(ms2, mc1.m);
            ms3 = _mm_madd_pi16(ms3, mc1.m);
            
            ms0 = _mm_add_pi32(ms0, ms2);
            ms1 = _mm_add_pi32(ms1, ms3);
            
            //
            ms4 = _mm_adds_pi16(*(src4+j), *(srca+j));
            ms3 = _mm_adds_pi16(*(src5+j), *(src9+j));
            ms2 = ms3;
            
            ms2 = _mm_unpackhi_pi16(ms2, ms4);
            ms3 = _mm_unpacklo_pi16(ms3, ms4);

            ms2 = _mm_madd_pi16(ms2, mc2.m);
            ms3 = _mm_madd_pi16(ms3, mc2.m);
            
            ms0 = _mm_add_pi32(ms0, ms2);
            ms1 = _mm_add_pi32(ms1, ms3);
            
            //
            ms4 = _mm_adds_pi16(*(src6+j), *(src8+j));
            ms3 = *(src7+j);
            ms2 = ms3;
            
            ms2 = _mm_unpackhi_pi16(ms2, ms4);
            ms3 = _mm_unpacklo_pi16(ms3, ms4);
            
            ms2 = _mm_madd_pi16(ms2, mc3.m);
            ms3 = _mm_madd_pi16(ms3, mc3.m);
            
            ms0 = _mm_add_pi32(ms0, ms2);
            ms1 = _mm_add_pi32(ms1, ms3);
            
            ms0 = _mm_srai_pi32(ms0, 15);
            ms1 = _mm_srai_pi32(ms1, 15);
            
            *(dst0+j) = _mm_packs_pi32(ms1, ms0);
        }
        
        dst += s;
        src += s;
    }

    _mm_empty();
}


const VLPF vflt_mmx[] = 
{ lpf_v_3_mmx,  lpf_v_5_mmx,  lpf_v_7_mmx,
  lpf_v_9_mmx, lpf_v_11_mmx, lpf_v_13_mmx, lpf_v_15_mmx, NULL };

